import json
import typing

from aio_pika import IncomingMessage
from marshmallow import Schema, types


def load_from_msg(
        schema: Schema,
        msg: IncomingMessage,
        *,
        many: typing.Optional[bool] = None,
        partial: typing.Optional[typing.Union[bool, types.StrSequenceOrSet]] = None,
        unknown: typing.Optional[str] = None,
        **kwargs
):
    """Deserialize bytes from message body via json to schema."""
    return schema.load(
        json.loads(msg.body),
        many=many,
        partial=partial,
        unknown=unknown,
        **kwargs,  # noqa
    )



import abc
import inspect
import json
import logging
from functools import partial

import aio_pika
import marshmallow
from aio_pika import Exchange, IncomingMessage, Message, types

from .http_status import HTTPStatus
from .response import Response

logger = logging.getLogger(__name__)

ACTION_DATA_KEY = '__action_data__'


def action(
        basename: str = None,
        durable: bool = False,
        exclusive: bool = False,
        timeout: types.TimeoutType = None,
        auto_delete: bool = False,
        arguments: dict = None,
        passive=False,
):
    if arguments is None:
        arguments = {}

    def decorator(func: callable):
        async def wrapper(consumer: AbstractConsumer, exchange: Exchange, message: IncomingMessage):
            try:
                return await func(consumer, exchange, message)
            except marshmallow.exceptions.ValidationError as err:
                logger.exception(
                    'Fail while processing incoming request #{0}'.format(
                        message.message_id))
                return await consumer.publish(exchange, message, Response.from_exc(
                    status=HTTPStatus.BAD_REQUEST,
                    exc=err,
                    data={
                        **err.normalized_messages()
                    }
                ))
            except Exception as err:
                logger.exception(
                    'Unable to process incoming request #{0}'.format(
                        message.message_id))
                return await consumer.publish(exchange, message, Response.from_exc(
                    status=HTTPStatus.INTERNAL_SERVER_ERROR,
                    exc=err,
                ))

        # TODO: x-dead-letter-exchange
        setattr(wrapper, ACTION_DATA_KEY, {
            'basename': basename or func.__name__,
            'durable': durable,
            'exclusive': exclusive,
            'timeout': timeout,
            'auto_delete': auto_delete,
            'arguments': arguments,
            'passive': passive,
        })
        return wrapper

    return decorator


class AbstractConsumer(abc.ABC):
    DEFAULT_ACTIONS = {
        'create',
        'retrieve',
        'list',
        'partial_update',
        'destroy',
    }

    consumer_key: str
    channel_number: int = 1
    prefetch_count: int = 100

    def __init__(self, connection: aio_pika.Connection):
        self.connection = connection
        self.channel = self.connection.channel(channel_number=self.channel_number)

    async def _create_queue(self, basename, **kwargs) -> aio_pika.Queue:
        """
        :param basename: the base to use for the queue name that are created.
        :return: created queue
        """
        return await self.channel.declare_queue(
            name=':'.join([self.consumer_key, basename]),
            **kwargs
        )

    async def initialize(self):
        """Initialize channel, queue and consumers."""
        await self.channel.initialize()
        await self.channel.set_qos(self.prefetch_count)

        consumers = []
        data: dict
        for name, method_ref in inspect.getmembers(self, predicate=inspect.ismethod):
            if data := getattr(method_ref, ACTION_DATA_KEY, None):
                queue = await self._create_queue(**data)

                # Start listening the queue
                await queue.consume(partial(
                    method_ref, self.channel.default_exchange)
                )
                consumers.append(queue)

        logger.debug('Registered {0} consumers of {1}: {2}'.format(
            len(consumers),
            self.__class__.__name__,
            consumers,
        ))

    @staticmethod
    async def publish(exchange: Exchange, message: IncomingMessage, response: Response):
        await exchange.publish(
            Message(
                headers=response.headers,
                body=json.dumps(response.data).encode(),
                content_type='application/json',
                correlation_id=message.correlation_id,
            ),
            routing_key=message.reply_to,
        )



from .abstract_consumer import AbstractConsumer, action
from .helpers import load_from_msg
from .http_status import HTTPStatus
from .producer import Producer
from .request import Request
from .response import Response

__all__ = (
    'AbstractConsumer',
    'action',
    'Producer',
    'Response',
    'Request',
    'HTTPStatus',
    # helpers
    'load_from_msg',
)



import traceback
from typing import Union

from amqp_framework.http_status import HTTPStatus


class Response:

    def __init__(self, status: Union[HTTPStatus, int], headers: dict = None, data: dict = None):

        if isinstance(status, HTTPStatus):
            self._status = status.value
        else:
            self._status = HTTPStatus(status)

        if headers is None:
            headers = {}
        else:
            assert isinstance(headers, dict), 'Unsupported type {0} for headers, dict required.' \
                .format(type(headers))
        self._headers = headers

        if data is None:
            data = {}
        else:
            assert isinstance(data, dict), 'Unsupported type {0} for data, dict required.' \
                .format(type(data))
        self.data = data

    def __repr__(self):
        return "<Response(status='{0}', data='{1}', headers='{2}')>".format(
            self._status, self.data, self._headers
        )

    @property
    def headers(self):
        return {
            **self._headers,
            'status': self._status,
        }

    @classmethod
    def from_exc(cls, status: Union[HTTPStatus, int], exc: Exception, headers=None, data=None):
        if not data:
            data = {}

        return cls(status=status, headers=headers, data={
            **data,
            **{
                'exc_info': getattr(exc, 'code', None) or status.phrase,
                'exc_traceback': traceback.format_exc(),
            },
        })



"""
Setup.py for module micomp.
"""

from setuptools import setup, find_packages  # Always prefer setuptools over distutils


def get_long_description():
    with open('README.md', 'r') as fh:
        return fh.read()


setup(
    name='amqp_framework',
    version='0.1.1',
    description='Awesome producers and consumers powered by aio_pika.',
    long_description=get_long_description(),
    license='MIT',

    # The project's main homepage.
    url='https://github.com/aobcvr/amqp_framework',

    # Author details
    author='aobcvr',
    author_email='aobcvr@gmail.com',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        'Programming Language :: Python :: 3',

        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.9',

    # What does your project relate to?
    keywords='amqp aio_pika',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),
    install_requires=[
        'aio-pika==6.8',
        'marshmallow',
        'environs',
    ],
    extras_require={
        'dev': [
            # Linters
            'isort==5.9.3',
            # Static analyse
            'flake8==3.9.2',
        ]
    }
)



class Request:

    def __init__(self, headers: dict = None, data: dict = None):
        if headers is None:
            headers = {}
        else:
            assert isinstance(headers, dict), 'Unsupported type {0} for headers, dict required.' \
                .format(type(headers))
        self.headers = headers

        if data is None:
            data = {}
        else:
            assert isinstance(data, dict), 'Unsupported type {0} for data, dict required.' \
                .format(type(data))
        self.data = data

    def __repr__(self):
        return "<Response(data='{0}', headers='{1}')>".format(
            self.data, self.headers
        )



import asyncio
import json
import uuid

import aio_pika
from environs import Env

from .response import Response

env = Env()
env.read_env()  # also reads the .env file in project root

AMQP_URI = env.str('AMQP_URI')


class Producer:
    def __init__(self, loop=None):
        self.connection = None
        self.channel = None
        self.callback_queue = None
        self.futures = {}
        self.loop = loop or asyncio.get_event_loop()

    async def connect(self):
        self.connection: aio_pika.Connection = await aio_pika.connect_robust(
            AMQP_URI, loop=self.loop
        )
        self.channel = await self.connection.channel()
        self.callback_queue = await self.channel.declare_queue(exclusive=True)
        await self.callback_queue.consume(self.on_response, timeout=5)

        return self

    def on_response(self, message: aio_pika.IncomingMessage):
        future = self.futures.pop(message.correlation_id)
        future.set_result(
            Response(
                status=message.headers.get('status'),
                headers=dict(message.headers),
                data=json.loads(message.body)
            )
        )

    async def call(self, routing_key: str, body=None) -> Response:
        correlation_id = str(uuid.uuid4())
        future = self.loop.create_future()

        self.futures[correlation_id] = future

        if not body:
            body = {}

        await self.channel.default_exchange.publish(
            aio_pika.Message(
                body=json.dumps(body).encode(),
                content_type="application/json",
                correlation_id=correlation_id,
                reply_to=self.callback_queue.name,
            ),
            routing_key=routing_key,
        )
        return await future
